import os

import requests
from tqdm import tqdm

from dataset_sh.utils.files import checksum


def direct_upload_to_url(presigned_url, file_path):
    with open(file_path, "rb") as f:
        response = requests.put(presigned_url, data=f)
        response.raise_for_status()


def multipart_upload_to_s3(
        instruction,
        file_path,
):
    """
    instruction:
    {
        'allowed': True,
        'action': 'upload',
        'parts': parts,
        'finished_parts': existing.finished,
        'cancel_url': cancel_url,
        'finish_url': finish_url
    }

    Args:
        instruction:
        file_path:
        silent:
        print_fn:

    Returns:

    """

    parts = instruction["parts"]
    finish_url = instruction["finish_url"]
    cancel_url = instruction["cancel_url"]

    uploaded = {p for p in instruction.get("finished_parts", [])}

    print(f'Starting multipart upload, in case you need to cancel, submit a post request to {cancel_url} ')

    with open(file_path, "rb") as f:
        total_parts = len(parts)
        for idx, part in enumerate(parts):

            idx_label = f"{idx + 1}/{total_parts}"

            """
            part = {
                    'part': part_number,
                    'start': chunk_start,
                    'size': chunk_end - chunk_start,
                    'url': create_part_upload_url(
                        hostname,
                        dataset_locator.namespace, dataset_locator.dataset_name,
                        checksum_value, file_size, part_number, token),
                    'report_url': create_part_progress_report_url(
                        hostname,
                        dataset_locator.namespace, dataset_locator.dataset_name,
                        checksum_value, file_size, part_number, token
                    )
            }
            """
            part_number = part["part"]
            start = part["start"]
            size = part["size"]
            upload_url = part["url"]
            report_url = part["report_url"]

            if part_number in uploaded:
                print(f"✅ {idx_label} - Skipping part id: {part_number}")
                continue

            f.seek(start)
            chunk = f.read(size)

            print(f"⬆️ {idx_label} Uploading part id {part_number} - Byte range: ({start}-{start + size})...")
            r = requests.put(upload_url, data=chunk)
            r.raise_for_status()

            # Mark part as uploaded
            resp = requests.post(report_url, json={
                "action": "progress-update",
                "part": part_number,
                "etag": r.headers.get("ETag"),
            })

    print("✅ Uploading done. Completing...")
    resp = requests.post(finish_url)
    resp.raise_for_status()
    print("🎉 Upload complete.")


class FileServerClient:
    endpoint: str
    headers: dict

    def __init__(self, endpoint: str, headers: dict = None):
        self.endpoint = endpoint
        self.headers = headers

    def upload(
            self,
            file: str,
            namespace: str,
            dataset_name: str,
            tags=None
    ):
        """

        Args:
            file:
            namespace:
            dataset_name:
            tags:

        Returns:

        """

        if not os.path.exists(file):
            raise FileNotFoundError(f"File not found: {file}")
        version = checksum(file)
        remote_filename = f'{namespace}/{dataset_name}:version={version}'
        instruction = self.request_upload(file, remote_filename, version, tags)
        if not instruction['allowed']:
            raise ValueError("Upload not allowed")

        multipart_upload_to_s3(
            instruction,
            file,
        )

        # if instruction['type'] == 'direct':
        #     # Direct upload to URL
        #     with open(file, 'rb') as f:
        #         response = requests.put(instruction['url'], data=f)
        #         response.raise_for_status()
        #     print("🎉 Upload complete.")
        #
        # elif instruction['type'] in ['multipart']:
        #     multipart_upload_to_s3(
        #         instruction,
        #         file,
        #     )
        # else:
        #     raise ValueError(f"Unsupported upload type: {instruction['type']}")

    def request_upload(self, file: str, remote_filename: str, checksum_value: str, tags=None):
        filesize = os.path.getsize(file)
        req = {
            'action': 'upload',
            'filename': remote_filename,
            'size': filesize,
            'checksum': checksum_value,
            'tags': tags,
        }

        resp = requests.post(self.endpoint, params=req, headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    def request_download(self, file: str):
        req = {
            'action': 'download',
            'file': file,
        }

        resp = requests.post(self.endpoint, params=req, headers=self.headers)
        resp.raise_for_status()
        return resp.json()



import math
from urllib.parse import urlencode

from dataset_sh.clients.locator import DatasetLocator
from dataset_sh.clients.remote import url_join


def create_download_url(hostname, locator, token):
    query_string = urlencode({
        "token": token
    })

    if locator.version is not None:
        p = f'api/dataset/{locator.namespace}/{locator.dataset_name}/version/{locator.version}/file?{query_string}'
    elif locator.tag is not None:
        p = f'api/dataset/{locator.namespace}/{locator.dataset_name}/tag/{locator.tag}/file?{query_string}'
    else:
        p = f'api/dataset/{locator.namespace}/{locator.dataset_name}/tag/latest/file?{query_string}'

    return url_join(hostname, p)


def get_chunk_size(
        file_size,
        min_chunk_size=5 * 1024 * 1024,
        max_chunks=10_000
):
    chunk_size = max(min_chunk_size, (file_size + max_chunks - 1) // max_chunks)
    chunk_count = int(math.ceil(file_size / chunk_size))
    return chunk_size, chunk_count


def suggest_parts(
        hostname,
        dataset_locator: DatasetLocator,
        file_size,
        checksum_value,
        token,
        min_chunk_size=5 * 1024 * 1024,
        max_chunks=10_000
):
    """
    calculate multipart upload parts based on the given parameters.
    :param file_size: size of the file in bytes
    :param min_chunk_size: minimal chunk size in bytes
    :param max_chunks: maximal number of chunks
    :return: list of part dicts {start: int, end: int, part: int}, the end is not inclusive
    """
    file_size = int(file_size)
    if file_size <= 0:
        return []

    chunk_size, chunk_count = get_chunk_size(file_size, min_chunk_size, max_chunks)

    parts = []

    for i in range(chunk_count):
        chunk_start = i * chunk_size
        chunk_end = min(chunk_start + chunk_size, file_size)
        part_number = i
        parts.append({
            'part': part_number,
            'start': chunk_start,
            'size': chunk_end - chunk_start,
            'url': create_part_upload_url(
                hostname,
                dataset_locator.namespace, dataset_locator.dataset_name,
                checksum_value, file_size, part_number, token),
            'report_url': create_part_progress_report_url(
                hostname,
                dataset_locator.namespace, dataset_locator.dataset_name,
                checksum_value, file_size, part_number, token
            )
        })

    return parts


def create_part_upload_url(hostname, namespace, dataset_name, checksum_value, file_size, part_number: int, token: str):
    query_string = urlencode({
        "token": token,
        "part_id": part_number,
        'checksum_value': checksum_value,
        'file_size': file_size,
    })

    p = f'api/dataset/{namespace}/{dataset_name}/upload?{query_string}'
    return url_join(hostname, p)


def create_part_progress_report_url(hostname, namespace, dataset_name, checksum_value, file_size, part_number: int,
                                    token: str):
    query_string = urlencode({
        "token": token,
        'action': 'progress-update',
        "part_id": part_number,
        'checksum_value': checksum_value,
        'file_size': file_size,
    })

    p = f'api/dataset/{namespace}/{dataset_name}/upload-progress?{query_string}'
    return url_join(hostname, p)


def create_part_finish_url(hostname, namespace, dataset_name, checksum_value, file_size, part_number: int, token: str):
    query_string = urlencode({
        "token": token,
        'action': 'done',
        "part_id": part_number,
        'checksum_value': checksum_value,
        'file_size': file_size,
    })

    p = f'api/dataset/{namespace}/{dataset_name}/upload-progress?{query_string}'
    return url_join(hostname, p)


def create_upload_finish_url(hostname, namespace, dataset_name, checksum_value, file_size, tags, token: str):
    query_string = urlencode({
        "token": token,
        'action': 'done',
        "tags": tags,
        'checksum_value': checksum_value,
        'file_size': file_size,
    })

    p = f'api/dataset/{namespace}/{dataset_name}/upload-progress?{query_string}'
    return url_join(hostname, p)


def create_upload_cancel_url(hostname, namespace, dataset_name, checksum_value, file_size, token: str):
    query_string = urlencode({
        "token": token,
        'action': 'abort',
        'checksum_value': checksum_value,
        'file_size': file_size,
    })

    p = f'api/dataset/{namespace}/{dataset_name}/upload-progress?{query_string}'
    return url_join(hostname, p)



import os.path

CONFIG_JSON = os.path.expanduser('~/.dataset_sh_config.json')
DEFAULT_AUTHOR_JSON = os.path.expanduser('~/.dataset_sh_author.json')

STORAGE_BASE = os.path.expanduser('~/dataset_sh/storage')
DAEMON_PID_FILE = os.environ.get(
    'DSH_DAEMON_PID_FILE',
    os.path.expanduser('~/dataset_sh/daemon.pid')
)

DSH_DEFAULT_HOST = os.getenv('DSH_DEFAULT_HOST', 'https://base.dataset.sh')

SAMPLE_CHAR_COUNT = 1000 * 10

DEFAULT_COLLECTION_NAME = 'main'



from dataclasses import dataclass

import click
import docker
import os
from datetime import datetime

from dataset_sh import Remote, LocalStorage
from dataset_sh.io import DatasetFileWriter
from dataset_sh.server.core import RepoServerConfig
from dataset_sh.utils.files import checksum
from tests.core.test_io import NameAndCount


def get_default_root_folder():
    """Generate default root folder path with current date."""
    current_date = datetime.now().strftime('%Y_%m_%d')
    return os.path.join('/tmp', 'dataset_sh_sys_testing', f'test_{current_date}')


@dataclass
class DatasetTestingBaseFolder:
    base: str

    @property
    def base_path(self):
        return self.base

    @property
    def server_config(self):
        return os.path.abspath(os.path.join(self.base, 'server.config'))

    @property
    def client_config(self):
        return os.path.abspath(os.path.join(self.base, 'client.config'))

    @property
    def data_dir(self):
        return os.path.abspath(os.path.join(self.base, 'data'))

    @property
    def uploader_dir(self):
        return os.path.abspath(os.path.join(self.base, 'uploader'))

    @property
    def posts_dir(self):
        return os.path.abspath(os.path.join(self.base, 'posts'))

    @property
    def download_dir(self):
        return os.path.abspath(os.path.join(self.base, 'download'))

    @property
    def draft_dir(self):
        return os.path.abspath(os.path.join(self.base, 'drafts'))

    def prepare(self):
        os.makedirs(self.base, exist_ok=True)
        os.makedirs(self.data_dir, exist_ok=True)
        os.makedirs(self.uploader_dir, exist_ok=True)
        os.makedirs(self.posts_dir, exist_ok=True)
        os.makedirs(self.download_dir, exist_ok=True)
        os.makedirs(self.draft_dir, exist_ok=True)


def start_server(
        root_folder: DatasetTestingBaseFolder,
        image_name='dsh-dev-local',
        port=29999,
):
    environment = {
        'DSH_APP_HOSTNAME': f'http://localhost:{port}',
    }

    volumes = {
        str(root_folder.data_dir): {
            'bind': '/app/data',
            'mode': 'rw'
        },
        str(root_folder.posts_dir): {
            'bind': '/app/posts',
            'mode': 'rw'
        },
        str(root_folder.uploader_dir): {
            'bind': '/app/uploader',
            'mode': 'rw'
        },
        str(root_folder.server_config): {
            'bind': '/app/dataset-sh-server-config.json',  # exact file path inside container
            'mode': 'ro'  # optional: read-only
        }
    }

    client = docker.from_env()

    try:
        container = client.containers.run(
            image_name,
            name='dsh-test-upload',
            detach=True,
            environment=environment,
            volumes=volumes,
            ports={'8989/tcp': port},
            restart_policy={"Name": "unless-stopped"},
        )

        print(f"Container started: {container.id}")

        # Print container logs
        for line in container.logs(stream=True):
            print(line.decode('utf-8').strip())

    except docker.errors.APIError as e:
        print(f"Error creating container: {e}")
    except docker.errors.ImageNotFound as e:
        print(f"Image not found: {e}")


def run_upload():
    pass


def run_download():
    pass


def is_same_file(f1, f2):
    return checksum(f1) == checksum(f2)


def run_all(root_dir: DatasetTestingBaseFolder):
    """

    1. start server
        1. create config with an user key
        2. mount to config to server container
        3. run container
    2. create draft dataset
    3. upload data to draft dataset
        1. check if data is uploaded
    4. download data from draft dataset
        1. check if data is downloaded

    """
    root_dir.prepare()

    username = 'test-user'
    password = 'test-password'
    host = 'http://localhost:29999'

    # create config with an user key
    repo_config = RepoServerConfig()
    repo_config.update_password(username, password)
    repo_config.allow_upload = True
    repo_config.write_to_file(root_dir.server_config)
    access_key = repo_config.generate_key(username, password)

    start_server(root_dir)

    dataset_file = os.path.join(root_dir.data_dir, 'test-upload.dataset')

    with DatasetFileWriter(dataset_file) as writer:
        writer.add_collection('main', [
            NameAndCount('a', i).to_dict() for i in range(100)
        ])

    version = checksum(dataset_file)

    remote = Remote(host=host, access_key=access_key)
    remote.dataset('test/test-upload').upload_from_file(dataset_file, ['latest'])


    dd = LocalStorage(root_dir.download_dir).dataset('test/test-download')
    remote.dataset('test/test-upload').latest().download_to(dd)


@click.group()
def cli():
    """Dataset.sh CLI tool for managing local development server and data transfer."""
    pass


@cli.command()
@click.option('--image-name', default='dsh37', help='Docker image name')
@click.option('--port', default=29999, type=int, help='Port to run the server on')
@click.option('--root-folder', default=None, help='Root folder for server data')
def server(image_name, port, root_folder):
    """Start the development server."""
    if root_folder is None:
        root_folder = get_default_root_folder()
    start_server(image_name=image_name, port=port, root_folder=root_folder)


@cli.command()
def upload():
    """Upload data to the server."""
    run_upload()


@cli.command()
def download():
    """Download data from the server."""
    run_download()


if __name__ == '__main__':
    cli()



import abc
import os
from typing import Optional
from importlib.metadata import version as lib_version

import requests


class MarkedFolder:
    location: str
    parent: Optional['MarkedFolder']

    def __init__(self, location, parent=None):
        self.location = location
        self.parent = parent

    @abc.abstractmethod
    def marker_file(self) -> str:
        raise NotImplementedError()

    def post_create(self):
        return

    def exists(self):
        """
        this marked folder exists iff its marker file exist.
        Returns:

        """
        return os.path.isfile(self.marker_file())

    def create_location(self):
        os.makedirs(self.location, exist_ok=True)

    def create(self):
        self.create_location()
        if not self.exists():
            with open(self.marker_file(), 'w') as out:
                out.write('1')
            self.post_create()
        if self.parent and not self.parent.exists():
            self.parent.create()

    def create_if_not_exists(self):
        if not self.exists():
            self.create()


def version_check(remote_server_url):
    """
    check dataset.sh version of a remote server.
    Returns:

    """
    local_version = lib_version('dataset.sh')
    ping_url = remote_server_url
    api_url = url_join(self.remote.host, 'api', 'info')
    headers = self.remote.get_headers()
    resp = requests.get(
        api_url,
        headers=headers,
    )
    resp.raise_for_status()
    return resp.json()


def compare_version(local_version, remote_version):
    pass



import datetime
from urllib.parse import urlencode

import jwt
from flask import request


def get_token(
        secret_key,
        action,
        filename,
        expiration_minutes=5
):
    # Add expiration time (exp) and issued at (iat) claims

    payload = {
        'action': action,
        'filename': filename
    }

    now = datetime.datetime.now(datetime.UTC)
    payload.update({
        'exp': now + datetime.timedelta(minutes=expiration_minutes),
        'iat': now
    })

    # Create the JWT token
    token = jwt.encode(
        payload=payload,
        key=secret_key,
        algorithm='HS256'
    )

    return token


def extract_token(token, secret_key):
    try:
        # Verify and decode the token
        payload = jwt.decode(
            jwt=token,
            key=secret_key,
            algorithms=['HS256']
        )
        return payload

    except jwt.ExpiredSignatureError:
        # Token has expired
        return None
    except jwt.InvalidTokenError:
        # Token is invalid
        return None


def extract_token_from_request(secret_key):
    token = request.args.get('token')
    return extract_token(token, secret_key)


def create_download_token(secret_key, filename):
    return get_token(secret_key, 'download', filename, expiration_minutes=10)


def verify_download_token(secret_key, filename):
    token = request.args.get('token')
    payload = extract_token(token, secret_key)
    if payload is None:
        return False
    if payload['action'] != 'download':
        return False
    if payload['filename'] != filename:
        return False
    return True


def create_upload_token(secret_key, filename):
    return get_token(secret_key, 'upload', filename, expiration_minutes=3600)


def verify_upload_token(secret_key, filename):
    token = request.args.get('token')
    payload = extract_token(token, secret_key)
    if payload is None:
        return False
    if payload['action'] != 'upload':
        return False
    if payload['filename'] != filename:
        return False
    return True
